/*
 *   MediaMVP Server
 *
 *   (C) 2003 Dominic Morris
 *
 *   $Id: menuapp.c,v 1.1.1.1 2003/12/04 00:10:13 dom Exp $
 *   $Date: 2003/12/04 00:10:13 $
 *
 *
 *   Implements the application
 */

#include "menuapp.h"
#include "libini.h"

#include <dirent.h>
#include <sys/stat.h>



typedef struct {
    char   *name;
    int     flags;
} filelist_t;


struct _menuapp {
    dongle_t  *dongle;
    render_t  *render;
    app_t     *app;
    menu_t    *menu;
    char      *current_media;
    int        num_next_media;
    char     **next_media;
    char       paused;
};


typedef struct {
    menuapp_t      *menuapp;
    char       *path;
    /* For handling scanning of directories */
    int         filelist_num;
    filelist_t *filelist;
} dirlist_t;



typedef struct {
    char     *name;
    char     *url;
} channel_t;

static char         *c_livetv_file = NULL;
static char         *c_liveradio_file = NULL;
static char         *c_mp3_dir     = ".";
static char         *c_mpeg_dir    = ".";

static int           livetv_num    = 0;
static channel_t    *livetv        = NULL;
static time_t        livetv_time   = 0;

static int           liveradio_num    = 0;
static channel_t    *liveradio        = NULL;
static time_t        liveradio_time   = 0;

static int           menuapp_main(menu_t *menu,void *imenuapp, int sel);
static int           menuapp_live_tv(menu_t *menu,void *imenuapp, int sel);
static int           process_livetv(menu_t *menu,void *imenuapp, int sel);
static int           livetv_keys(void *ptr, int code);
static int           livetv_ack(int acktype, void *param, unsigned char *buf, int len);

static int           menuapp_live_radio(menu_t *menu,void *imenuapp, int sel);
static int           process_liveradio(menu_t *menu,void *imenuapp, int sel);
static int           liveradio_keys(void *ptr, int code);
static int           liveradio_ack(int acktype, void *param, unsigned char *buf, int len);


static int           menuapp_mpeg(menu_t *menu,void *imenuapp, int sel);
static int           process_mpeg(menu_t *menu,void *imenuapp, int sel);
static int           mpeg_keys(void *ptr, int code);
static int           mpeg_ack(int acktype, void *param, unsigned char *buf, int len);
static int           mpeg_colour_keys(menu_t *menu, void *imenuapp, int key, int sel);

static int           menuapp_music(menu_t *menu,void *imenuapp, int sel);
static int           process_music(menu_t *menu,void *imenuapp, int sel);
static int           music_keys(void *ptr, int code);
static int           music_ack(int acktype, void *param, unsigned char *buf, int len);
static int           music_colour_keys(menu_t *menu, void *imenuapp, int key, int sel);


static int           menuapp_settings(menu_t *menu,void *imenuapp, int sel);
static int           process_settings(menu_t *menu,void *imenuapp, int sel);

static filelist_t   *directory_scan(char *path, int *scanned);
static void          directory_sort(int num,filelist_t *list);

static dirlist_t    *new_dirlist(menuapp_t *menuapp);
static void          delete_dirlist(void *dir);    /* Ahem */


static void          next_media_clear(menuapp_t *menuapp);
static void          next_media_pop(menuapp_t *menuapp);
static void          next_media_push(menuapp_t *menuapp, char *url);



static int           read_channel_file(char *filename, int *num, channel_t **list, time_t *update_time);

void menuapp_config()
{
    iniparse_add("media:livetv",OPT_STR,&c_livetv_file);
    iniparse_add("media:liveradio",OPT_STR,&c_liveradio_file);
    iniparse_add("media:mp3",OPT_STR,&c_mp3_dir);
    iniparse_add("media:mpeg",OPT_STR,&c_mpeg_dir);
}


void menuapp_init()
{
    if ( c_livetv_file != NULL ) {
        if ( read_channel_file(c_livetv_file,&livetv_num,&livetv,&livetv_time)  < 0 ) {
            printf("Malformed file %s\n",c_livetv_file);
            exit(1);
        }
    }   
    if ( c_liveradio_file != NULL ) {
        if ( read_channel_file(c_liveradio_file,&liveradio_num,&liveradio,&liveradio_time)  < 0 ) {
            printf("Malformed file %s\n",c_liveradio_file);
            exit(1);
        }
    }   

}

menuapp_t *new_menuapp(dongle_t *dongle, render_t *render, app_t *app)
{
    menuapp_t      *menuapp = malloc(sizeof(*menuapp));

    menuapp->render = render;
    menuapp->app    = app;
    menuapp->dongle = dongle;
    menuapp->current_media = NULL;
    menuapp->num_next_media = 0;
    menuapp->next_media = NULL;
    menuapp->paused = FALSE;


    menuapp_main(NULL,menuapp,0);    

    return menuapp;
}


void delete_menuapp(menuapp_t *menuapp)
{

    delete_menu_tree(menuapp->menu);

    if ( menuapp->current_media ) {
        free(menuapp->current_media);
        menuapp->current_media = NULL;
    }

    if ( menuapp->next_media ) {
        next_media_clear(menuapp);
    }

    free(menuapp);
}




static int menuapp_main(menu_t *menu,void *imenuapp, int sel)
{   
    menuapp_t   *menuapp = imenuapp;

    menuapp->menu = new_menu(menuapp->dongle,menuapp->render,menuapp->app,NULL,menuapp,MENU_NUMBERS);
    menu_set_title(menuapp->menu,"MediaMVP");
    menu_add(menuapp->menu,"Live TV",menuapp_live_tv);
    menu_add(menuapp->menu,"Live Radio",menuapp_live_radio);
    menu_add(menuapp->menu,"Recordings",menuapp_mpeg);
    menu_add(menuapp->menu,"Music",menuapp_music);
    menu_add(menuapp->menu,"Settings",menuapp_settings);
    menu_display(menuapp->menu);
    return 1;
}


/***********************
 *
 *   Handling of the Live TV Menus
 *
 ***********************/

static int menuapp_live_tv(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t     *menuapp = imenuapp;
    channel_t *chan;
    menu_t    *child;
    int        i;

    /* Force a re-read of the channel file */
    read_channel_file(c_livetv_file,&livetv_num,&livetv,&livetv_time);

    child = new_menu(menuapp->dongle,menuapp->render,menuapp->app,menu,menuapp,0);
    menu_set_title(child,"Live TV"); 

    for ( i = 0; i < livetv_num; i++ ) {
        chan = &livetv[i];
        menu_add(child,chan->name,process_livetv);
    }
    menu_display(child);
    return 1;
}

static int process_livetv(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t     *menuapp = imenuapp;

    channel_t *chan = &livetv[sel];
    gui_send_play(menuapp->dongle->gui,chan->url);
    app_register_keys(menuapp->app,REGISTER_MEDIA,livetv_keys,menuapp);
    app_register_ack(menuapp->app,REGISTER_MEDIA,livetv_ack,menuapp);
    return 1;
}

static int livetv_ack(int acktype, void *param, unsigned char *buf, int len)
{
    menuapp_t     *menuapp = param;

    switch ( acktype ) {
    case RDC_PLAY:
        if ( (stream_get_type(menuapp->dongle->stream) & MEDIA_MASK) == MEDIA_MPEG ) {
            gui_send_message(menuapp->dongle->gui,&rfb_media_menu);
            gui_send_message(menuapp->dongle->gui,&rfb_display_off);
        }
        break;
    case RDC_STOP:             
        /* The client has sent a stop message, this could be response to "Stop" key press or
           us sending out a stop so we can switch to the net media file */ 
        if ( menuapp->num_next_media ) {
            next_media_pop(menuapp);           
            menuapp->paused = FALSE;
            gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        } else {
            if ( ( stream_get_lasttype(menuapp->dongle->stream) & MEDIA_MASK ) == MEDIA_MPEG ) {
                gui_send_message(menuapp->dongle->gui,&rfb_media_menu);
                gui_send_message(menuapp->dongle->gui,&rfb_display_on);
            }
            if ( menuapp->current_media ) {
                free(menuapp->current_media);
                menuapp->current_media = NULL;
            }
            menuapp->paused = FALSE;
            /* Remove registered handlers */
            app_register_keys(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
            app_register_ack(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
        }


        break;
    default:
        return 0;    /* Kick upstairs to menu ack handler */
    }

    return 1;       /* No need to pass back to menu handler */
}


static int livetv_keys(void *ptr, int code)
{
    menuapp_t    *menuapp = ptr;

    switch ( code ) {
    case kMenu:
    case kStop:
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
        menu_display(menuapp->menu);
        break;
    case kVolUp:
        gui_send_message(menuapp->dongle->gui,&rfb_media_volup);
        break;
    case kVolDn:
        gui_send_message(menuapp->dongle->gui,&rfb_media_voldown);
        break;
    case kMute:
        gui_send_message(menuapp->dongle->gui,&rfb_media_mute);
        break;
    default:
        return 0;  /* Do more keys */
    }
    return 1;
}


/***********************
 *
 *   Handling of the Live Radio Menus
 *
 ***********************/

static int menuapp_live_radio(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t     *menuapp = imenuapp;
    channel_t *chan;
    menu_t    *child;
    int        i;

    /* Force a re-read of the channel file */
    read_channel_file(c_liveradio_file,&liveradio_num,&liveradio,&liveradio_time);
    child = new_menu(menuapp->dongle,menuapp->render,menuapp->app,menu,menuapp,0);
    menu_set_title(child,"Live Radio Stations");  
    for ( i = 0; i < liveradio_num; i++ ) {
        chan = &liveradio[i];
        menu_add(child,chan->name,process_liveradio);
    } 
    menu_display(child);
    return 1;
}

static int process_liveradio(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t     *menuapp = imenuapp;
    channel_t *chan = &liveradio[sel];

    printf("Sending media do request %s\n",chan->url);
    if ( menuapp->current_media == NULL ) {        
        menuapp->current_media = strdup(chan->url);
        menuapp->paused = FALSE;
        gui_send_play(menuapp->dongle->gui,chan->url);
    } else {
        next_media_clear(menuapp);
        next_media_push(menuapp,chan->url);       
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
    }
    app_register_keys(menuapp->app,REGISTER_MEDIA,liveradio_keys,menuapp);
    app_register_ack(menuapp->app,REGISTER_MEDIA,liveradio_ack,menuapp);
    return 1;
}


static int liveradio_keys(void *ptr, int code)
{
    menuapp_t     *menuapp = ptr;

    switch ( code ) { 
    case kStop:
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
        next_media_clear(menuapp);
        menu_display(menuapp->menu);
        break;
    case kVolUp:
        gui_send_message(menuapp->dongle->gui,&rfb_media_volup);
        break;
    case kVolDn:
        gui_send_message(menuapp->dongle->gui,&rfb_media_voldown);
        break;
    case kMute:
        gui_send_message(menuapp->dongle->gui,&rfb_media_mute);
        break;
    default:
        return 0;      /* Carry on processing menu keys */
    }   
    return 1;
}

static int liveradio_ack(int acktype, void *param, unsigned char *buf, int len)
{
    menuapp_t     *menuapp = param;

    switch ( acktype ) {
    case RDC_STOP:     /* The client has sent a stop message */  
        if ( menuapp->num_next_media ) {
            next_media_pop(menuapp);
            menuapp->paused = FALSE;
            gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        } else {
            /* Remove registered handlers */
            if ( menuapp->current_media ) {
                free(menuapp->current_media);
                menuapp->current_media = NULL;
            }
            menuapp->paused = FALSE;
            app_register_keys(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
            app_register_ack(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
        }
        break;
    default:
        return 0;    /* Kick it up a level */
    }

    return 1;        /* Don't process anymore */
}


/*
 * Handle menus for mpeg files 
 */
static int menuapp_mpeg(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t      *menuapp = imenuapp;
    dirlist_t  *dirlist;
    menu_t     *child;
    char        buf[256];
    int         i;

    dirlist = new_dirlist(menuapp);

    dirlist->path = strdup(c_mpeg_dir);

    /* Scan the dirlist */
    dirlist->filelist = directory_scan(c_mpeg_dir,&dirlist->filelist_num);

    child = new_menu(menuapp->dongle,menuapp->render,menuapp->app,menu,dirlist,0);
    menu_set_title(child,"Video Recordings");   
    menu_set_param_clearup(child,delete_dirlist);

    for ( i = 0; i < dirlist->filelist_num; i++ ) {
        if ( dirlist->filelist[i].flags ) {
            snprintf(buf,sizeof(buf),"[%s]",dirlist->filelist[i].name);
        } else {
            snprintf(buf,sizeof(buf),"%s",dirlist->filelist[i].name);
        }       
        menu_add(child,buf,process_mpeg);
    }
    menu_set_colour_actions(child,"Back",NULL,NULL,NULL,mpeg_colour_keys);

    menu_display(child);
    return 1;  
}

static int process_mpeg(menu_t *menu,void *imenuapp, int sel)
{
    char         buf[FILENAME_MAX+1];
    dirlist_t    *dirlist = imenuapp;
    dirlist_t    *dirlist2;
    filelist_t   *entry;
    menu_t       *child;
    int           i;


    entry = &dirlist->filelist[sel];

    if ( entry->flags ) {
        /* The selection was a directory - we need to create a new menu - hah! */
        dirlist2 = new_dirlist(dirlist->menuapp);

        /* Scan the dirlist */
        snprintf(buf,sizeof(buf),"%s/%s",dirlist->path,entry->name);
        dirlist2->path = strdup(buf);
        dirlist2->filelist = directory_scan(buf,&dirlist2->filelist_num);

        child = new_menu(dirlist2->menuapp->dongle,dirlist2->menuapp->render,dirlist2->menuapp->app,menu,dirlist2,0);
        snprintf(buf,sizeof(buf),"Video - %s",entry->name);
        menu_set_title(child,buf);   
        menu_set_param_clearup(child,delete_dirlist);
            
        for ( i = 0; i < dirlist2->filelist_num; i++ ) {
            if ( dirlist2->filelist[i].flags ) {
                snprintf(buf,sizeof(buf),"[%s]",dirlist2->filelist[i].name);
            } else {
                snprintf(buf,sizeof(buf),"%s",dirlist2->filelist[i].name);
            }       
            menu_add(child,buf,process_mpeg);
        }
        menu_set_colour_actions(child,"Back",NULL,NULL,NULL,music_colour_keys);

        menu_display(child);
        return 1;        
    } else {
        snprintf(buf,sizeof(buf),"file://%s/%s",dirlist->path,entry->name);
        if ( dirlist->menuapp->current_media == NULL ) {        
            dirlist->menuapp->current_media = strdup(buf);
            dirlist->menuapp->paused = FALSE;
//            next_media_push(dirlist->menuapp,buf);
            gui_send_play(dirlist->menuapp->dongle->gui,buf);
        } else {
            next_media_clear(dirlist->menuapp);
            next_media_push(dirlist->menuapp,buf);           
            dirlist->menuapp->paused = FALSE;
            gui_send_message(dirlist->menuapp->dongle->gui,&rfb_media_stop);    
        }
        app_register_keys(dirlist->menuapp->app,REGISTER_MEDIA,mpeg_keys,dirlist->menuapp);
        app_register_ack(dirlist->menuapp->app,REGISTER_MEDIA,mpeg_ack,dirlist->menuapp);
    }  

    return 1;     /* Carry on displaying this menu */
}

static int mpeg_colour_keys(menu_t *menu, void *imenuapp, int key, int sel)
{
    switch ( key ) {
    case kRed:
        return 0;   
    default:
    }

    return 1;     /* Carry on displaying this menu */
}

static int mpeg_keys(void *ptr, int code)
{
    menuapp_t     *menuapp = ptr;

    switch ( code ) { 
    case kStop:
        next_media_clear(menuapp);
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
        menu_display(menuapp->menu);
        break;
    case kPlay:
        gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        menuapp->paused = FALSE;
        break;
    case kPause:
        if ( menuapp->paused ) {
            menuapp->paused = FALSE;
            gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        } else {
            menuapp->paused = TRUE;
            gui_send_message(menuapp->dongle->gui,&rfb_media_pause);
        }
        break;
    case kVolUp:
        gui_send_message(menuapp->dongle->gui,&rfb_media_volup);
        break;
    case kVolDn:
        gui_send_message(menuapp->dongle->gui,&rfb_media_voldown);
        break;
    case kMute:
        gui_send_message(menuapp->dongle->gui,&rfb_media_mute);
        break;
#if 0
        /* Forward/reverse don't work...erk! */
    case kFastFwd:
        gui_send_message(menuapp->dongle->gui,&rfb_media_forward);
        break;
    case kFastRew:
        gui_send_message(menuapp->dongle->gui,&rfb_media_rewind);
        break;
#endif
    case kSkip:
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
        break;
    default:
        return 0;      /* Carry on processing menu keys */
    }   
    return 1;
}


static int mpeg_ack(int acktype, void *param, unsigned char *buf, int len)
{
    menuapp_t         *menuapp = param;

    switch ( acktype ) {
    case RDC_PLAY:
        if ( (stream_get_type(menuapp->dongle->stream) & MEDIA_MASK) == MEDIA_MPEG ) {
            gui_send_message(menuapp->dongle->gui,&rfb_media_menu);
            gui_send_message(menuapp->dongle->gui,&rfb_display_off);
        }
        break;
    case RDC_STOP:     /* The client has sent a stop message */ 
        if ( menuapp->num_next_media ) {
            next_media_pop(menuapp);
            menuapp->paused = FALSE;
            gui_send_play(menuapp->dongle->gui,menuapp->current_media);          
        } else {
          if ( ( stream_get_lasttype(menuapp->dongle->stream) & MEDIA_MASK ) == MEDIA_MPEG ) {
                gui_send_message(menuapp->dongle->gui,&rfb_media_menu);
                gui_send_message(menuapp->dongle->gui,&rfb_display_on);
            }
            if ( menuapp->current_media ) {
                free(menuapp->current_media);
                menuapp->current_media = NULL;
            }
            menuapp->paused = FALSE;
            /* Remove registered handlers */
            app_register_keys(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
            app_register_ack(menuapp->app,REGISTER_MEDIA,NULL,menuapp);            
        }
        break;
    case RDC_FORWARD:
    case RDC_REWIND:
        gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        break;
    default:
        return 0;    /* Kick it up a level */
    }

    return 1;        /* Don't process anymore */
}




/*
 *   Handle the root menu for Music menu
 */
static int menuapp_music(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t      *menuapp = imenuapp;
    dirlist_t  *dirlist;
    menu_t     *child;
    char        buf[256];
    int         i;

    dirlist = new_dirlist(menuapp);

    dirlist->path = strdup(c_mp3_dir);

    /* Scan the dirlist */
    dirlist->filelist = directory_scan(c_mp3_dir,&dirlist->filelist_num);

    child = new_menu(menuapp->dongle,menuapp->render,menuapp->app,menu,dirlist,0);
    menu_set_title(child,"Music");   
    menu_set_param_clearup(child,delete_dirlist);

    for ( i = 0; i < dirlist->filelist_num; i++ ) {
        if ( dirlist->filelist[i].flags ) {
            snprintf(buf,sizeof(buf),"[%s]",dirlist->filelist[i].name);
        } else {
            snprintf(buf,sizeof(buf),"%s",dirlist->filelist[i].name);
        }       
        menu_add(child,buf,process_music);
    }
    menu_set_colour_actions(child,"Back","Play dir",NULL,NULL,music_colour_keys);

    menu_display(child);
    return 1;
}


static int process_music(menu_t *menu,void *imenuapp, int sel)
{
    char         buf[FILENAME_MAX+1];
    dirlist_t    *dirlist = imenuapp;
    dirlist_t    *dirlist2;
    filelist_t   *entry;
    menu_t       *child;
    int           i;


    entry = &dirlist->filelist[sel];

    if ( entry->flags ) {
        /* The selection was a directory - we need to create a new menu - hah! */
        dirlist2 = new_dirlist(dirlist->menuapp);

        /* Scan the dirlist */
        snprintf(buf,sizeof(buf),"%s/%s",dirlist->path,entry->name);
        dirlist2->path = strdup(buf);
        dirlist2->filelist = directory_scan(buf,&dirlist2->filelist_num);

        child = new_menu(dirlist2->menuapp->dongle,dirlist2->menuapp->render,dirlist2->menuapp->app,menu,dirlist2,0);
        snprintf(buf,sizeof(buf),"Music - %s",entry->name);
        menu_set_title(child,buf);   
        menu_set_param_clearup(child,delete_dirlist);
            
        for ( i = 0; i < dirlist2->filelist_num; i++ ) {
            if ( dirlist2->filelist[i].flags ) {
                snprintf(buf,sizeof(buf),"[%s]",dirlist2->filelist[i].name);
            } else {
                snprintf(buf,sizeof(buf),"%s",dirlist2->filelist[i].name);
            }       
            menu_add(child,buf,process_music);
        }
        menu_set_colour_actions(child,"Back","Play dir",NULL,NULL,music_colour_keys);

        menu_display(child);
        return 1;        
    } else {
        snprintf(buf,sizeof(buf),"file://%s/%s",dirlist->path,entry->name);
        if ( dirlist->menuapp->current_media == NULL ) {        
            dirlist->menuapp->current_media = strdup(buf);
            dirlist->menuapp->paused = FALSE;
            gui_send_play(dirlist->menuapp->dongle->gui,buf);
        } else {
            next_media_clear(dirlist->menuapp);
            next_media_push(dirlist->menuapp,buf);           
            dirlist->menuapp->paused = FALSE;
            gui_send_message(dirlist->menuapp->dongle->gui,&rfb_media_stop);    
        }
        app_register_keys(dirlist->menuapp->app,REGISTER_MEDIA,music_keys,dirlist->menuapp);
        app_register_ack(dirlist->menuapp->app,REGISTER_MEDIA,music_ack,dirlist->menuapp);
    }  

    return 1;     /* Carry on displaying this menu */
}

static int music_colour_keys(menu_t *menu, void *imenuapp, int key, int sel)
{
    char           buf[FILENAME_MAX+1];
    dirlist_t     *dir = imenuapp;
    int            i;
    char           play = TRUE;

    switch ( key ) {
    case kRed:
        return 0;
    case kGreen:       /* Add all contents of the directory */
        if ( dir->menuapp->current_media ) {
            play = FALSE;
        }
        next_media_clear(dir->menuapp);

        for ( i = 0; i < dir->filelist_num; i++ ) {
            if ( dir->filelist[i].flags == 0 ) {
                snprintf(buf,sizeof(buf),"file://%s/%s",dir->path,dir->filelist[i].name);
                next_media_push(dir->menuapp,buf);
            }
        }
        if ( dir->menuapp->num_next_media ) {
            if ( play ) {
                next_media_pop(dir->menuapp);
                gui_send_play(dir->menuapp->dongle->gui,dir->menuapp->current_media);
                app_register_keys(dir->menuapp->app,REGISTER_MEDIA,music_keys,dir->menuapp);
                app_register_ack(dir->menuapp->app,REGISTER_MEDIA,music_ack,dir->menuapp);
            } else {
                gui_send_message(dir->menuapp->dongle->gui,&rfb_media_stop);
            }
        }
        break;
    default:
    }

    return 1;     /* Carry on displaying this menu */
}

static int music_keys(void *ptr, int code)
{
    menuapp_t     *menuapp = ptr;

    switch ( code ) { 
    case kStop:
        next_media_clear(menuapp);
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
        menu_display(menuapp->menu);
        break;
    case kPlay:
        gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        menuapp->paused = FALSE;
        break;
    case kPause:
        if ( menuapp->paused ) {
            menuapp->paused = FALSE;
            gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        } else {
            menuapp->paused = TRUE;
            gui_send_message(menuapp->dongle->gui,&rfb_media_pause);
        }
        break;
    case kVolUp:
        gui_send_message(menuapp->dongle->gui,&rfb_media_volup);
        break;
    case kVolDn:
        gui_send_message(menuapp->dongle->gui,&rfb_media_voldown);
        break;
    case kMute:
        gui_send_message(menuapp->dongle->gui,&rfb_media_mute);
        break;
    case kFastFwd:
        gui_send_message(menuapp->dongle->gui,&rfb_media_forward);
        break;
    case kFastRew:
        gui_send_message(menuapp->dongle->gui,&rfb_media_rewind);
        break;
    case kSkip:
        gui_send_message(menuapp->dongle->gui,&rfb_media_stop);
        break;
    default:
        return 0;      /* Carry on processing menu keys */
    }   
    return 1;
}

static int music_ack(int acktype, void *param, unsigned char *buf, int len)
{
    menuapp_t         *menuapp = param;

    switch ( acktype ) {
    case RDC_STOP:     /* The client has sent a stop message */  
        if ( menuapp->num_next_media ) {
            next_media_pop(menuapp);
            menuapp->paused = FALSE;
            gui_send_play(menuapp->dongle->gui,menuapp->current_media);          
        } else {
            /* Remove registered handlers */
            if ( menuapp->current_media ) {
                free(menuapp->current_media);
                menuapp->current_media = NULL;
            }
            menuapp->paused = FALSE;
            app_register_keys(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
            app_register_ack(menuapp->app,REGISTER_MEDIA,NULL,menuapp);
        }
        break;
    case RDC_FORWARD:
    case RDC_REWIND:
        gui_send_play(menuapp->dongle->gui,menuapp->current_media);
        break;
    default:
        return 0;    /* Kick it up a level */
    }

    return 1;        /* Don't process anymore */
}




/***********************
 *
 *   Handling of the settings Menu
 *
 ***********************/
static int menuapp_settings(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t           *menuapp = imenuapp;
    menu_t          *child;
    static char     *tvmode[]  = { "NTSC", "PAL" };
    static char     *flicker[] = { "None", "Low", "Medium", "High" };
    static char     *output[]  = { "RGB Detected", "Svideo", "Composite", "RGB/Composite" };
    static char     *aspect[]  = { "4:3", "16:9" };

    child = new_menu(menuapp->dongle,menuapp->render,menuapp->app,menu,menuapp,0);
    menu_set_title(child,"Settings");

 

    /* Add in all possible options - we'll validate things on the callback */
    menu_add_option(child,"TV Mode:",process_settings,
                    MENU_INT,2,tvmode,&menuapp->dongle->tvmode);

    /* A little bit of fiddling - see rules in proces_settings */
    if ( menuapp->dongle->videooutput != 0 ) {
        menu_add_option(child,"Output:",process_settings,
                        MENU_INT,4,output,&menuapp->dongle->videooutput);
    } else {
        menu_add(child,"Output: RGB Detected",process_settings);
    }

    menu_add_option(child,"Flicker Control:",process_settings,
                    MENU_INT,4,flicker,&menuapp->dongle->flickermode);

    menu_add_option(child,"Aspect Ratio:",process_settings,
                    MENU_INT,2,aspect,&menuapp->dongle->aspectratio);

    menu_add(child,"Cancel changes",process_settings);
    menu_add(child,"Save settings (may take a while)",process_settings);

    menu_display(child);
    return 1;
}

static int process_settings(menu_t *menu,void *imenuapp, int sel)
{
    menuapp_t   *menuapp = imenuapp;
    /* Some rules preened from the configuration.html:

    1). For NTSC unit -- in response to GUI query, video output value 
    is always '3', so you show "Video Output" as "Composite/Svideo" no changes
    allowed !!

    2). For SCART Unit --
    a). '0' : then you show "Video Output" as "RGB Detected", No changes 
    allowed !!!
    (b). '1' : then you show "Video Output" as "SVIDEO", it can only be
    changed to RGB/COMPOSITE (value == 2)
    (c). '2': you show "Video Output" as "RGB/COMPOSITE", it can only be 
    changed to SVIDEO (value == 1)

    We try to obey them [but probably fail miserable]!
    */

    /* If we come through with -1 then we should test the settings */
    switch ( sel ) {
    case -1:  /* If we come through with -1 then we should test the settings */
        if ( menuapp->dongle->tvmode == 0 ) {
            menuapp->dongle->videooutput = 3;
        }
        gui_send_settings(menuapp->dongle->gui,RDC_SETTINGS_TEST);
        return 1;  /* Carry on displaying current menu */
    case 4:
        gui_send_settings(menuapp->dongle->gui,RDC_SETTINGS_CANCEL);
        return 0;
    case 5:
        gui_send_settings(menuapp->dongle->gui,RDC_SETTINGS_SAVE);
        return 0;
    default:
        return 0;   /* Carry on displaying */
    } 
}



static dirlist_t    *new_dirlist(menuapp_t *menuapp)
{
    dirlist_t   *dir = malloc(sizeof(*dir));

    dir->filelist_num = 0;
    dir->filelist = NULL;
    dir->menuapp = menuapp;

    return dir;
}

static void          delete_dirlist(void *idir)
{
    dirlist_t *dir = idir;
    int        i;

    for ( i = 0; i < dir->filelist_num; i++ ) {
        free(dir->filelist[i].name);
    }
    if ( dir->filelist ) {
        free(dir->filelist);
    }

    free(dir->path);

    free(dir);
}

static filelist_t *directory_scan(char *path, int *scanned)
{
    DIR    *dir;
    struct dirent *dirent;
    struct stat sb;
    char   filename[FILENAME_MAX+1];
    filelist_t *list = NULL;
    int         num  = 0;
    int         flags;
    int         i;


    *scanned = 0;

    if ( ( dir = opendir(path) ) == NULL ) {
        perror("opendir");
        return NULL;
    }

    while ( (dirent = readdir(dir)) != NULL ) {
        if ( dirent->d_name[0] == '.' ) {
            continue;
        }

        snprintf(filename,sizeof(filename),"%s/%s",path,dirent->d_name);
        if ( stat(filename,&sb) < 0 ) {
            continue;
        }

        if ( S_ISDIR(sb.st_mode) ) {
            flags = 1;
        } else {
            flags = 0;
        }

        i = num++;
        list = realloc(list, num * sizeof(filelist_t));

        list[i].name = strdup(dirent->d_name);
        list[i].flags = flags;
            
    }

    closedir(dir);

    directory_sort(num,list);

    *scanned = num;
    return list;
}

static void directory_sort(int num,filelist_t *list)
{
    int    i,j;
    int    t_flags;
    char  *t_name;

    for ( i = (num-1) ; i >= 0; i-- ) {
        t_name = list[i].name;
        t_flags = list[i].flags;
        j = i + 1;

        while ( j < num && strcasecmp(list[j].name,t_name) < 0 ) {
            list[j-1].name = list[j].name;
            list[j-1].flags = list[j].flags;

            j++;
        }

        list[j-1].name = t_name;
        list[j-1].flags = t_flags;

    }
}



static int read_channel_file(char *filename, int *num, channel_t **list, 
                             time_t *update_time)
{
    char         buf[500];
    int          channum;
    channel_t   *chan = *list;
    char        *ptr;
    FILE        *fp;
    struct stat  sb;
    int          i;

    if ( stat(filename,&sb) < 0 ) {
        return 0;
    }

    if ( *update_time < sb.st_mtime ) {
        *update_time = sb.st_mtime;
        for ( i = 0; i < *num; i++ ) {
            free(chan[i].name);
            free(chan[i].url);
        }


        if ( ( fp = fopen(filename,"r") ) == NULL ) {
            return -1;
        }

        if ( chan != NULL ) {
            free(chan);
        }
        chan = NULL;
        channum = 0;

        *list = NULL;
        *num = 0;

        while ( !feof(fp) && fgets(buf,sizeof(buf),fp) ) {
            if ( (ptr = strchr(buf,';') ) == NULL ) {
                continue;
            }
            *ptr = 0;
            ptr++;

            i = channum++;

            chan = realloc(chan,channum * sizeof(channel_t));
            chan[i].name = strdup(buf);
            chan[i].url = strdup(ptr);            
        }
        *list = chan;
        *num = channum;
    }

    return 0;
}

static void next_media_clear(menuapp_t *menuapp)
{
    int         i;

    if ( menuapp->num_next_media ) {
        for ( i = 0; i > menuapp->num_next_media; i++ ) {
            free(menuapp->next_media[i]);
        }
        free(menuapp->next_media);
        menuapp->next_media = NULL;
        menuapp->num_next_media = 0;
    }
}

static void next_media_pop(menuapp_t *menuapp)
{
    int       i;

    if ( menuapp->num_next_media ) {     /* Which it already is... */
        if ( menuapp->current_media ) {
            free(menuapp->current_media);
        }

        menuapp->current_media = menuapp->next_media[0];

        for ( i = 1; i < menuapp->num_next_media; i++ ) {
            menuapp->next_media[i-1] = menuapp->next_media[i];
        }
        if ( --menuapp->num_next_media == 0 ) {
            free(menuapp->next_media);
            menuapp->next_media = NULL;
        }
    } 
}

static void next_media_push(menuapp_t *menuapp, char *url)
{
    int       i = menuapp->num_next_media++;

    menuapp->next_media = realloc(menuapp->next_media,menuapp->num_next_media * sizeof(char *));
    menuapp->next_media[i] = strdup(url);
}



/*
 * Local Variables:
 *  indent-tabs-mode:nil
 *  require-final-newline:t
 *  c-basic-offset: 4
 * End:
 */
